"""Passive BLE monitor integration."""
import asyncio
from threading import Thread

import aioblescan as aiobs
from datetime import datetime

from bleparser import BleParser

from fhempy.lib.ble_monitor.bt_helpers import reset_bluetooth, BT_INTERFACES

from .const import (
    CONF_DEVICES,
    CONF_ACTIVE_SCAN,
    CONF_BT_AUTO_RESTART,
    CONF_BT_INTERFACE,
    CONF_DEVICE_ENCRYPTION_KEY,
    CONF_DEVICE_TRACK,
    CONF_HCI_INTERFACE,
    CONF_GATEWAY_ID,
    CONF_REPORT_UNKNOWN,
    CONF_DISCOVERY,
    MEASUREMENT_DICT,
)

from .helper import dict_get_or_clean, dict_get_or, identifier_clean


class HCIdump(Thread):
    """Mimic deprecated hcidump tool."""

    def __init__(self, logger, config, dataqueue):
        """Initiate HCIdump thread."""
        Thread.__init__(self)
        self.logger = logger
        self.logger.debug("HCIdump thread: Init")
        self.dataqueue_bin = dataqueue["binary"]
        self.dataqueue_meas = dataqueue["measuring"]
        self.dataqueue_tracker = dataqueue["tracker"]
        self._event_loop = None
        self._joining = False
        self.evt_cnt = 0
        self.config = config
        self._interfaces = config[CONF_HCI_INTERFACE]
        self._active = int(config[CONF_ACTIVE_SCAN] is True)
        self.discovery = True
        self.filter_duplicates = True
        self.aeskeys = {}
        self.sensor_whitelist = []
        self.tracker_whitelist = []
        self.report_unknown = False
        self.last_bt_reset = datetime.now()
        if self.config[CONF_REPORT_UNKNOWN]:
            self.report_unknown = self.config[CONF_REPORT_UNKNOWN]
            self.logger.info(
                "Attention! Option report_unknown is enabled for %s sensors, "
                "be ready for a huge output",
                self.report_unknown,
            )
        # prepare device:key lists to speedup parser
        if self.config[CONF_DEVICES]:
            for device in self.config[CONF_DEVICES]:
                if (
                    CONF_DEVICE_ENCRYPTION_KEY in device
                    and device[CONF_DEVICE_ENCRYPTION_KEY]
                ):
                    p_id = bytes.fromhex(dict_get_or_clean(device).lower())
                    p_key = bytes.fromhex(device[CONF_DEVICE_ENCRYPTION_KEY].lower())
                    self.aeskeys[p_id] = p_key
                else:
                    continue
        self.logger.debug("%s encryptors mac:key pairs loaded", len(self.aeskeys))

        # prepare sensor whitelist to speedup parser
        if (
            isinstance(self.config[CONF_DISCOVERY], bool)
            and self.config[CONF_DISCOVERY] is False
        ):
            self.discovery = False
            if self.config[CONF_DEVICES]:
                for device in self.config[CONF_DEVICES]:
                    self.sensor_whitelist.append(dict_get_or(device))

        # remove duplicates from sensor whitelist
        self.sensor_whitelist = list(dict.fromkeys(self.sensor_whitelist))
        self.logger.debug(
            "sensor whitelist: [%s]", ", ".join(self.sensor_whitelist).upper()
        )
        for i, key in enumerate(self.sensor_whitelist):
            self.sensor_whitelist[i] = bytes.fromhex(identifier_clean(key))
        self.logger.debug(
            "%s sensor whitelist item(s) loaded", len(self.sensor_whitelist)
        )

        # prepare device tracker list to speedup parser
        if self.config[CONF_DEVICES]:
            for device in self.config[CONF_DEVICES]:
                if CONF_DEVICE_TRACK in device and device[CONF_DEVICE_TRACK]:
                    track_key = bytes.fromhex(dict_get_or_clean(device))
                    self.tracker_whitelist.append(track_key)
                else:
                    continue
        self.logger.debug(
            "%s device tracker(s) being monitored", len(self.tracker_whitelist)
        )

        # prepare the ble_parser
        self.ble_parser = BleParser(
            report_unknown=self.report_unknown,
            discovery=self.discovery,
            filter_duplicates=self.filter_duplicates,
            sensor_whitelist=self.sensor_whitelist,
            tracker_whitelist=self.tracker_whitelist,
            aeskeys=self.aeskeys,
        )

    def process_hci_events(self, data, gateway_id=""):
        """Parse HCI events."""
        self.evt_cnt += 1
        if len(data) < 12:
            return
        sensor_msg, tracker_msg = self.ble_parser.parse_data(data)
        if sensor_msg:
            measurements = list(sensor_msg.keys())
            device_type = sensor_msg["type"]
            sensor_list = (
                MEASUREMENT_DICT[device_type][0] + MEASUREMENT_DICT[device_type][1]
            )
            binary_list = MEASUREMENT_DICT[device_type][2] + ["battery"]
            measuring = any(x in measurements for x in sensor_list)
            binary = any(x in measurements for x in binary_list)
            if binary == measuring:
                self.dataqueue_bin.sync_q.put_nowait(sensor_msg)
                self.dataqueue_meas.sync_q.put_nowait(sensor_msg)
            else:
                if binary is True:
                    self.dataqueue_bin.sync_q.put_nowait(sensor_msg)
                if measuring is True:
                    self.dataqueue_meas.sync_q.put_nowait(sensor_msg)
        if tracker_msg:
            tracker_msg[CONF_GATEWAY_ID] = gateway_id
            self.dataqueue_tracker.sync_q.put_nowait(tracker_msg)

    def run(self):
        """Run HCIdump thread."""
        while True:
            self.logger.debug("HCIdump thread: Run")
            mysocket = {}
            fac = {}
            conn = {}
            btctrl = {}
            interface_is_ok = {}
            interfaces_to_reset = []
            initialized_evt = {}
            if self._event_loop is None:
                self._event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._event_loop)
            if "disable" not in self.config[CONF_BT_INTERFACE]:
                for hci in self._interfaces:
                    interface_is_ok[hci] = False
                    try:
                        mysocket[hci] = aiobs.create_bt_socket(hci)
                    except OSError as error:
                        self.logger.error(
                            "HCIdump thread: OS error (hci%i): %s", hci, error
                        )
                    else:
                        fac[hci] = getattr(
                            self._event_loop, "_create_connection_transport"
                        )(mysocket[hci], aiobs.BLEScanRequester, None, None)
                        conn[hci], btctrl[hci] = self._event_loop.run_until_complete(
                            fac[hci]
                        )
                        # Wait up to five seconds for aioblescan BLEScanRequester to initialize
                        initialized_evt[hci] = getattr(btctrl[hci], "_initialized")
                        self.logger.debug(
                            "HCIdump thread: BLEScanRequester._initialized is %s for hci%i, "
                            " waiting for connection...",
                            initialized_evt[hci].is_set(),
                            hci,
                        )
                        try:
                            self._event_loop.run_until_complete(
                                asyncio.wait_for(initialized_evt[hci].wait(), 5)
                            )
                        except asyncio.TimeoutError:
                            self.logger.error(
                                "HCIdump thread: Something wrong - interface hci%i not ready,"
                                " and will be skipped for current scan period.",
                                hci,
                            )
                        else:
                            btctrl[hci].process = self.process_hci_events
                            self.logger.debug("HCIdump thread: connected to hci%i", hci)
                            try:
                                self._event_loop.run_until_complete(
                                    btctrl[hci].send_scan_request(self._active)
                                )
                            except RuntimeError as error:
                                self.logger.error(
                                    "HCIdump thread: Runtime error while sending scan request on hci%i: %s.",
                                    hci,
                                    error,
                                )
                            else:
                                interface_is_ok[hci] = True
                                self.logger.debug(
                                    "HCIdump thread: BLEScanRequester._initialized is %s for hci%i, "
                                    " connection established, send_scan_request succeeded.",
                                    initialized_evt[hci].is_set(),
                                    hci,
                                )
                    if (interface_is_ok[hci] is False) and (
                        self.config[CONF_BT_AUTO_RESTART] is True
                    ):
                        interfaces_to_reset.append(hci)
                if interfaces_to_reset:
                    ts_now = datetime.now()
                    if (ts_now - self.last_bt_reset).seconds > 60:
                        for iface in interfaces_to_reset:
                            self.logger.error(
                                "HCIdump thread: Trying to power cycle Bluetooth adapter hci%i %s,"
                                " will try to use it next scan period.",
                                iface,
                                BT_INTERFACES[iface],
                            )
                            reset_bluetooth(iface)
                        self.last_bt_reset = ts_now
            self.logger.debug("HCIdump thread: start main event_loop")
            try:
                self._event_loop.run_forever()
            finally:
                self.logger.debug("HCIdump thread: main event_loop stopped, finishing.")
                if "disable" not in self.config[CONF_BT_INTERFACE]:
                    for hci in self._interfaces:
                        if interface_is_ok[hci] is True:
                            try:
                                self._event_loop.run_until_complete(
                                    btctrl[hci].stop_scan_request()
                                )
                            except RuntimeError as error:
                                self.logger.error(
                                    "HCIdump thread: Runtime error while stop scan request on hci%i: %s.",
                                    hci,
                                    error,
                                )
                            except KeyError:
                                self.logger.debug(
                                    "HCIdump thread: Key error while stop scan request on hci%i",
                                    hci,
                                )
                        try:
                            conn[hci].close()
                        except KeyError:
                            self.logger.debug(
                                "HCIdump thread: Key error while closing connection on hci%i",
                                hci,
                            )
                self._event_loop.run_until_complete(asyncio.sleep(0))
            if self._joining is True:
                break
            self.logger.debug("HCIdump thread: Scanning will be restarted")
            self.logger.debug(
                "%i HCI events processed for previous period", self.evt_cnt
            )
            self.evt_cnt = 0
        self._event_loop.close()
        self.logger.debug("HCIdump thread: Run finished")

    def join(self, timeout=1):
        """Join HCIdump thread."""
        self.logger.debug("HCIdump thread: joining")
        self._joining = True
        try:
            self._event_loop.call_soon_threadsafe(self._event_loop.stop)
        except AttributeError as error:
            self.logger.debug("%s", error)
        finally:
            Thread.join(self, timeout)
            self.logger.debug("HCIdump thread: joined")

    def restart(self):
        """Restarting scanner."""
        try:
            self._event_loop.call_soon_threadsafe(self._event_loop.stop)
        except AttributeError as error:
            self.logger.debug("%s", error)



import binascii
import ctypes
import logging
import struct
import time
from collections import namedtuple
from datetime import datetime
from uuid import UUID

import pygatt
from pygatt.exceptions import BLEError, NotConnectedError, NotificationTimeout

_LOGGER = logging.getLogger("nespresso_ble")

# Use full UUID since we do not use UUID from bluepy.btle
CHAR_UUID_DEVICE_NAME = UUID("00002a00-0000-1000-8000-00805f9b34fb")
CHAR_UUID_MANUFACTURER_NAME = UUID("00002a00-0000-1000-8000-00805f9b34fb")
CHAR_UUID_STATE = UUID("06aa3a12-f22a-11e3-9daa-0002a5d5c51b")
CHAR_UUID_NBCAPS = UUID("06aa3a15-f22a-11e3-9daa-0002a5d5c51b")
CHAR_UUID_SLIDER = UUID("06aa3a22-f22a-11e3-9daa-0002a5d5c51b")
CHAR_UUID_WATER_HARDNESS = UUID("06aa3a44-f22a-11e3-9daa-0002a5d5c51b")
CHAR_UUID_AUTH = UUID("06aa3a41-f22a-11e3-9daa-0002a5d5c51b")

Characteristic = namedtuple("Characteristic", ["uuid", "name", "format"])

c_uint8 = ctypes.c_uint8

manufacturer_characteristics = Characteristic(
    CHAR_UUID_MANUFACTURER_NAME, "manufacturer", "utf-8"
)
device_info_characteristics = [
    manufacturer_characteristics,
    Characteristic(CHAR_UUID_DEVICE_NAME, "device_name", "utf-8"),
]


class Flags_bits(ctypes.LittleEndianStructure):
    _fields_ = [
        ("bit0", c_uint8, 1),  # asByte & 1
        ("bit1", c_uint8, 1),  # asByte & 2
        ("bit2", c_uint8, 1),  # asByte & 4
        ("bit3", c_uint8, 1),  # asByte & 8
        ("bit4", c_uint8, 1),  # asByte & 16
        ("bit5", c_uint8, 1),  # asByte & 32
        ("bit6", c_uint8, 1),  # asByte & 64
        ("bit7", c_uint8, 1),  # asByte & 128
    ]


class Flags(ctypes.Union):
    _anonymous_ = ("bit",)
    _fields_ = [("bit", Flags_bits), ("asByte", c_uint8)]


class NespressoDeviceInfo:
    def __init__(self, manufacturer="", serial_nr="", model_nr="", device_name=""):
        self.manufacturer = manufacturer
        self.serial_nr = serial_nr
        self.model_nr = model_nr
        self.device_name = device_name

    def __str__(self):
        return "Manufacturer: {} Model: {} Serial: {} Device:{}".format(
            self.manufacturer, self.model_nr, self.serial_nr, self.device_name
        )


BYTE = Flags()
sensors_characteristics_uuid = [
    CHAR_UUID_STATE,
    CHAR_UUID_NBCAPS,
    CHAR_UUID_SLIDER,
    CHAR_UUID_WATER_HARDNESS,
]

sensors_characteristics_uuid_str = [str(x) for x in sensors_characteristics_uuid]


class BaseDecode:
    def __init__(self, name, format_type):
        self.name = name
        self.format_type = format_type

    def decode_data(self, raw_data):
        # val = struct.unpack(self.format_type,raw_data)
        val = raw_data
        if self.format_type == "caps_number":
            res = int.from_bytes(val, byteorder="big")
        elif self.format_type == "water_hardness":
            res = int.from_bytes(val[2:3], byteorder="big")
        elif self.format_type == "slider":
            res = binascii.hexlify(val)
            if (res) == b"00":
                res = 0
                # res = 'on'
            elif (res) == b"02":
                res = 1
                # res = 'off'
            else:
                res = "N/A"
        elif self.format_type == "state":
            BYTE0 = Flags()
            BYTE1 = Flags()
            BYTE2 = Flags()
            BYTE3 = Flags()

            BYTE0.asByte = val[0]
            BYTE1.asByte = val[1]
            # TODO error counter
            BYTE2.asByte = val[2]
            BYTE3.asByte = val[3]
            try:
                descaling_counter = int.from_bytes(val[6:9], byteorder="big")
            except Exception:
                _LOGGER.debug("can't get descaling counter")
                descaling_counter = 0
            return {
                "water_is_empty": BYTE0.bit0,
                "descaling_needed": BYTE0.bit2,
                "capsule_mechanism_jammed": BYTE0.bit4,
                "always_1": BYTE0.bit6,
                "water_temp_low": BYTE1.bit0,
                "awake": BYTE1.bit1,
                "water_engadged": BYTE1.bit2,
                "sleeping": BYTE1.bit3,
                "tray_sensor_during_brewing": BYTE1.bit4,
                "tray_open_tray_sensor_full": BYTE1.bit6,
                "capsule_engaged": BYTE1.bit7,
                "Fault": BYTE3.bit5,
                "descaling_counter": descaling_counter,
            }
        else:
            _LOGGER.debug("state_decoder else")
            res = val
        return {self.name: res}


sensor_decoders = {
    str(CHAR_UUID_STATE): BaseDecode(name="state", format_type="state"),
    str(CHAR_UUID_NBCAPS): BaseDecode(name="caps_number", format_type="caps_number"),
    str(CHAR_UUID_SLIDER): BaseDecode(name="slider", format_type="slider"),
    str(CHAR_UUID_WATER_HARDNESS): BaseDecode(
        name="water_hardness", format_type="water_hardness"
    ),
}


class NespressoDetect:
    def __init__(self, AUTH_CODE=None, mac=None):
        self.adapter = pygatt.backends.GATTToolBackend()
        self.nespresso_devices = [] if mac is None else [mac]
        self.auth_code = AUTH_CODE
        self.sensors = []
        self.sensordata = {}
        self.keep_connected = False
        self.dev = None

    def set_keep_connected(self, keep_connected):
        self.keep_connected = keep_connected

    def get_info(self, tries=0):
        # Try to get some info from the discovered Nespresso devices
        self.devices = {}

        for mac in self.nespresso_devices:
            device = NespressoDeviceInfo(serial_nr=mac)
            try:
                if self.dev is None:
                    self.adapter.start(reset_on_start=False)
                    self.dev = self.adapter.connect(
                        mac,
                        address_type=pygatt.BLEAddressType.random,
                        timeout=20,
                        auto_reconnect=self.keep_connected,
                    )
                for characteristic in device_info_characteristics:
                    try:
                        data = self.dev.char_read(characteristic.uuid)
                        setattr(
                            device,
                            characteristic.name,
                            data.decode(characteristic.format),
                        )
                    except (BLEError, NotConnectedError, NotificationTimeout):
                        self.dev = None
                        _LOGGER.error("Failed to read characteristics")
                if self.keep_connected is False:
                    self.dev = None
                    self.dev.disconnect()
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.debug("Failed to connect")
                time.sleep(5)  # wait 5s
                if tries < 3:
                    _LOGGER.debug("Failed to get_info, 3 times")
                    self.get_info(tries + 1)  # retry
                else:
                    _LOGGER.error("Failed to get_info, more than 3 times")
            if self.keep_connected is False:
                self.adapter.stop()
            self.devices[mac] = device

        return self.devices

    def get_sensors(self):
        self.sensors = {}
        for mac in self.nespresso_devices:
            try:
                if self.dev is None:
                    self.adapter.start(reset_on_start=False)
                    self.dev = self.adapter.connect(
                        mac,
                        address_type=pygatt.BLEAddressType.random,
                        timeout=20,
                        auto_reconnect=self.keep_connected,
                    )
                characteristics = self.dev.discover_characteristics()
                sensor_characteristics = []
                for characteristic in characteristics.values():
                    _LOGGER.debug(characteristic)
                    if characteristic.uuid in sensors_characteristics_uuid_str:
                        sensor_characteristics.append(characteristic)
                self.sensors[mac] = sensor_characteristics
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.error("Failed to discover sensors")
            if self.keep_connected is False:
                self.dev = None
                self.dev.disconnect()
                self.adapter.stop()

        return self.sensors

    def make_coffee(self, mac, temp="high", volume="lungo"):
        try:
            _LOGGER.debug("make flow a coffee")
            if self.dev is None:
                self.adapter.start(reset_on_start=False)
                self.dev = self.adapter.connect(
                    mac,
                    address_type=pygatt.BLEAddressType.random,
                    timeout=20,
                    auto_reconnect=self.keep_connected,
                )
            self.connectnespresso(self.dev)
            try:
                characteristic = "06aa3a42-f22a-11e3-9daa-0002a5d5c51b"
                command = "0305070400000000"
                if temp == "high":
                    command += "02"
                elif temp == "low":
                    command += "01"
                else:
                    command += "00"

                if volume == "ristretto":
                    command += "00"
                elif volume == "espresso":
                    command += "01"
                elif volume == "lungo":
                    command += "02"
                elif volume == "hotwater":
                    command += "04"
                elif volume == "americano":
                    command += "05"
                else:
                    command += "00"
                self.dev.char_write(
                    characteristic, binascii.unhexlify(command), wait_for_response=True
                )
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.error("Failed to write characteristic for coffee flow")
            if self.keep_connected is False:
                self.dev = None
                self.dev.disconnect()
        except (BLEError, NotConnectedError, NotificationTimeout):
            self.dev = None
            _LOGGER.error("Failed to connect for coffee flow")
        if self.keep_connected is False:
            self.adapter.stop()

    def connectnespresso(self, device, tries=0):
        try:
            # Write the auth code from android or Ios apps to the specific UUID to allow catching value from the machine
            device.char_write(
                CHAR_UUID_AUTH,
                binascii.unhexlify(self.auth_code),
                wait_for_response=True,
            )
        except (BLEError, NotConnectedError, NotificationTimeout):
            self.dev = None
            _LOGGER.debug("Failed to send auth code, retrying in 5s")
            time.sleep(5)  # wait 5s
            if tries < 3:
                _LOGGER.debug("Failed to send auth code, 3 times")
                self.connectnespresso(device, tries + 1)  # retry
            else:
                _LOGGER.error("Failed to send auth code, more than 3 times")

    def get_sensor_data(self):
        for mac, characteristics in self.sensors.items():
            try:
                if self.dev is None:
                    self.adapter.start(reset_on_start=False)
                    self.dev = self.adapter.connect(
                        mac,
                        address_type=pygatt.BLEAddressType.random,
                        timeout=20,
                        auto_reconnect=self.keep_connected,
                    )
                self.connectnespresso(self.dev)
                for characteristic in characteristics:
                    _LOGGER.debug("characteristic {}".format(characteristic))
                    try:
                        data = self.dev.char_read_handle(
                            "0x{:04x}".format(characteristic.handle)
                        )
                        if characteristic.uuid in sensor_decoders:
                            _LOGGER.debug(
                                "{} data {}".format(characteristic.uuid, data)
                            )
                            sensor_data = sensor_decoders[
                                characteristic.uuid
                            ].decode_data(data)
                            # sensor_data = str(data)
                            _LOGGER.debug(
                                "{} Got sensordata {}".format(mac, sensor_data)
                            )
                            if self.sensordata.get(mac) is None:
                                self.sensordata[mac] = sensor_data
                            else:
                                self.sensordata[mac].update(sensor_data)
                    except (BLEError, NotConnectedError, NotificationTimeout):
                        self.dev = None
                        _LOGGER.error("Failed to read characteristic")

                if self.keep_connected is False:
                    self.dev = None
                    self.dev.disconnect()
            except (BLEError, NotConnectedError, NotificationTimeout):
                self.dev = None
                _LOGGER.error("Failed to connect")
            if self.keep_connected is False:
                self.adapter.stop()

        return self.sensordata
